% ...
check_constraint_system(T, Prop, SimpleCS) :-
	get_constraint_system(T, Prop, CS),
	solution_exists(CS, SimpleCS).
check_leaf_constraints(T, Prop) :-
	get_mtree(T, MT),
	a23b([MT], [Prop], C),
	cc(C,CC),
	change_format(CC,CCC),
	solution_exists(CCC, _).
check_only_branch_constraints(T) :-
	get_mtree(T, MT),
	max_depth(MT, MD),
	iter(MT, 0, MD, NL),
	constraints_from_plain_list(NL, C),
	cc(C,CC),
	change_format(CC, CCC),
	solution_exists(CCC, _).

% get_constraint_system(+Tree, +Properties, -CS) :- -CS is the constraint system induced by mtree-structure +Tree;
%													+Properties are given to check for derivability by the solver later on.
get_constraint_system(T, Prop, FinalCS) :-
	get_mtree(T, MT),
	max_depth(MT, MD),
	iter(MT, 0, MD, NL),
	constraints_from_plain_list(NL, CS1),
	a23b([MT], [Prop], CS2),
	append(CS1, CS2, C),
	cc(C, CC),
	change_format(CC, FinalCS).
% max_depth(+Tree, ?D) :- unify D with the length of the longest path found in +Tree.
max_depth(Tree, D) :-
	findall(X, depth(Tree, _Node, 0, X), XL),
	max_list(XL, D).
% depth(+Tree, +Node, +CurrentDepthLevel, +MaxDepthLevel)
depth(sst(I,M,S,Ty,Pi,PIK,L,R,Ch), sst(I,M,S,Ty,Pi,PIK,L,R,Ch), N, N).
depth(sst(_,_,_,_,_,_,_,_,Ch), Node, N, D) :-
	member(Y, Ch),
	N1 is N + 1,
	depth(Y, Node, N1, D).
% iter(+Tree, +ThisDepth, +MaxDepth, ?ListOfNodes) :- iterate over all depth levels of +Tree, create a successive ?ListOfNodes with all nodes.
iter(T, I, D, Result) :-
	I = D,
	setof(Node, subnode_with_depth(T, I, Node), Result).
iter(T, I, D, Result) :-
	I < D,
	setof(Node, subnode_with_depth(T, I, Node), NL),
	I1 is I + 1,
	iter(T, I1, D, Tmp),
	append(NL, Tmp, Result).
% iter_list(+ListOfTrees, ?ListOfNodes) :- doing the same as iter, only for a list of trees and with respect to each tree's maxDepth.
iter_list([], []).
iter_list([T|Ts], ResultList) :-
	max_depth(T, D),
	iter(T, 0, D, R),
	iter_list(Ts, Rs),
	append(R, Rs, ResultList).
% swd(+Tree, +Depth, ?Node) :- ?Node is a node within +Tree with depth +Depth.
subnode_with_depth(sst(I,M,S,Ty,Pi,PIK,L,R,Ch), 0, sst(I,M,S,Ty,Pi,PIK,L,R,Ch)).
subnode_with_depth(sst(_,_,_,_,_,_,_,_,Ch), 1, Node) :- !,
	member(Node, Ch).
subnode_with_depth(sst(_,_,_,_,_,_,_,_,Ch), N, Node) :-
	N1 is N - 1,
	member(X, Ch),
	subnode_with_depth(X, N1, Node).

% constraints from plain list
constraints_from_plain_list([], []) :- !.
constraints_from_plain_list([sst(_,_,_,intruder,_,IK,LHS,_,_)|Ns], [constraint(LHS, IK)|Cs]) :-
	constraints_from_plain_list(Ns, Cs).
constraints_from_plain_list([sst(_,_,_,sc,_,IK,LHS,_,_)|Ns], [constraint(LHS, IK)|Cs]) :-
	constraints_from_plain_list(Ns, Cs).
constraints_from_plain_list([sst(_,_,_,Ty,_,_,_,_,_)|Ns], Cs) :-
	Ty \== sc,
	Ty \== intruder,
	constraints_from_plain_list(Ns, Cs).

% constraints from leaf
a23b([], _, []).
a23b([T|Ts], [P|Ps], CCCS) :-
	a23b_tree(T, P, CS),
	a23b(Ts, Ps, CCS),
	append(CS, CCS, CCCS).
a23b_tree(T, P, CS) :-
	leafs(T,L),
	a23b_leafs(L, P, CS).
a23b_leafs([], _, []).
a23b_leafs([sst(_,1,state(_,IK,_),_,_,_,_,_,_)|Ls], Propl, CCCS) :-
	a23b_leafs(Ls, Propl, CS),
	a23b_props(Propl, IK, CCS),
	append(CS, CCS, CCCS).
a23b_leafs([sst(_,0,_,_,_,_,_,_,_)|Ls], Propl, CS) :-
	a23b_leafs(Ls, Propl, CS).

a23b_props([], _, []).
a23b_props([P|Ps], IK, PropConstraints) :-
	a23b_props(Ps, IK, PCs),
	append([constraint(P, IK)], PCs, PropConstraints).

% our constraint format -> [[Var,Terms],...]
cc([],[]).
cc([constraint(T,L)|Cs], R) :-
	cc(Cs, TmpR),
	append([[T,L]], TmpR, R).

% change_format(+ConstraintList, ?NewConstraintList) :-	switch all items from +ConstraintList to the format expected in solver.pl to receive ?NewConstraintList.
change_format([], []) :- !.
change_format([C|Cs], [CC|CCs]) :-
	change_format_single(C, CC),
	change_format(Cs, CCs), !.
% termlist
change_format_termlist([], []) :- !.
change_format_termlist([T|Ts], [TT|TTs]) :-
	change_format_term(T, TT),
	change_format_termlist(Ts, TTs).
% single constraint
change_format_single([Term, IK], [CFTerm, CFIK]) :-
	change_format_term(Term, CFTerm),
	change_format_termlist(IK, CFIK).
% term
change_format_term(V, V) :- var(V), !.
change_format_term(X, X) :-
	atomic(X), !.
change_format_term(asym_enc(X, K), XX*KK) :-
	change_format_term(X, XX),
	change_format_term(K, KK).
change_format_term(sym_enc(X, K), XX+KK) :-
	change_format_term(X, XX),
	change_format_term(K, KK).
change_format_term(sig(X, K), XX/KK) :-
	change_format_term(X, XX),
	change_format_term(K, KK).
change_format_term(hash(X), sha(XX)) :-
	change_format_term(X, XX).
%change_format_term(var(X), var(X), [var(X)]).
%change_format_term(var(X,Y), var(X,Y), [var(X,Y)]).
change_format_term(TermList, CFTermList) :-
	is_list(TermList), !,
	change_format_termlist(TermList, CFTermList).
change_format_term(ComposedTerm, CFComposed) :-
	compound(ComposedTerm), !,
	ComposedTerm =.. [Head|Tail],	% do nothing for Head, assume atomic
	change_format_termlist(Tail, NTail),
	CFComposed =.. [Head|NTail].